# -*- coding: utf-8 -*-
"""
| Author: Alexandre CARRE 
| Created on: Jan 14, 2021
"""
import os
import warnings
from typing import List
from typing import Tuple, Union, Sequence

import SimpleITK as sitk
import numpy as np
import pandas as pd
import umap
from sklearn.manifold import TSNE
from sklearn.preprocessing import StandardScaler


def check_exist_vars(df: pd.DataFrame, _vars: List) -> np.ndarray:
    """
    Check that a list of columns name exist in a DataFrame.

    :param df: a DataFrame
    :param _vars: List of columns name to check
    :return: index of columns name
    :raise: ValueError if missing features
    """
    column_index = get_column_index(df, _vars)
    is_feature_present = column_index != -1
    if not isinstance(_vars, np.ndarray):
        _vars = np.array(_vars)
    if not is_feature_present.all():
        raise ValueError(f"Missing features: {', '.join(_vars[~is_feature_present].astype(str))}")
    return column_index


def get_column_index(df: pd.DataFrame, query_cols: List[str]) -> Union[np.ndarray]:
    """
    Get columns index from columns name

    :param df: input dataframe
    :param query_cols: List name of colunns
    :return: array of column index
    """
    return df.columns.get_indexer(query_cols)


def column_var_dtype(df: pd.DataFrame, identify_dtypes: Sequence[str] = ("object",)) -> pd.DataFrame:
    """
    identify type of columns in DataFrame

    :param df: input dataframe
    :param identify_dtypes: pandas dtype

    .. note::
        see `<https://pandas.pydata.org/pandas-docs/stable/user_guide/basics.html#basics-dtypes>`_ for pandas dtype

    :return: summary df with col index and col name for all identify_dtypes vars
    """
    col_type = df.dtypes
    col_names = list(df)

    cat_var_index = [i for i, x in enumerate(col_type) if x in identify_dtypes]
    cat_var_name = [x for i, x in enumerate(col_names) if i in cat_var_index]

    cat_var_df = pd.DataFrame({"ind": cat_var_index,
                               "name": cat_var_name})

    return cat_var_df


def one_hot_encoder(df: pd.DataFrame, columns: List[str], drop_column: bool = True, dummy_na: bool = False,
                    add_nan_columns: bool = False, inplace: bool = False) -> pd.DataFrame:
    """
    Encoding categorical feature in the dataframe, allow possibility to keep NaN.
    The categorical feature index and name are from cat_var function. These columns need to be "object" dtypes.

    :param df: input dataframe
    :param columns: List of columns to encode
    :param drop_column: Set to True to drop the original column after encoding. Default to True.
    :param dummy_na: Add a column to indicate NaNs, if False NaNs are ignored.
    :param add_nan_columns: Add a empty nan columns if not create (can be used are other categories)
    :param inplace: If False, return a copy. Otherwise, do operation inplace and return None
    :return: new dataframe where columns are one hot encoded
    """

    check_exist_vars(df, columns)

    if not inplace:
        df = df.copy(deep=True)

    for col in columns:
        dummies = pd.get_dummies(df[col], prefix=col, drop_first=False, dummy_na=dummy_na)
        df = pd.concat([df, dummies], axis=1)
        if add_nan_columns:
            if not any('nan' in x for x in dummies.columns):
                df[f'{col}_nan'] = 0
        if drop_column:
            df = df.drop(col, axis=1)

    return df


def fix_columns(df: pd.DataFrame, columns: List[str], inplace: bool = False, extra_nans: bool = False) -> pd.DataFrame:
    """
    Fix columns for the test set. When the train was encoded with :py:obj:`pd.get_dummies`.

    .. note::
        inspired from: `<http://fastml.com/how-to-use-pd-dot-get-dummies-with-the-test-set>`_

    :param df: input dataframe
    :param columns: columns of the original dataframe
    :param inplace: If False, return a copy. Otherwise, do operation inplace and return None
    :param extra_nans: put extra columns as nans based on one hot encoding columns
    :return: the corrected version of DataFrame for test set
    """
    if not inplace:
        df = df.copy(deep=True)

    missing_cols = set(columns) - set(df.columns)
    for c in missing_cols:
        df[c] = 0

    # make sure we have all the columns we need
    assert (set(columns) - set(df.columns) == set())

    extra_cols = set(df.columns) - set(columns)
    if extra_cols:
        warnings.warn(f"extra columns: {list(extra_cols)}")
        if extra_nans:
            all_nan_columns = [x.split('_nan')[0] for x in columns if 'nan' in x]
            for col in list(extra_cols):
                matching = [nan_col for nan_col in all_nan_columns if nan_col in col]
                if len(matching) != 1:
                    raise ValueError(f"matching was not found for adding extra columns in one hot encoding {matching}")
                df[f"{matching[0]}_nan"] += df[col]

    df = df[columns]
    return df


def scaler_encoder(df: pd.DataFrame, columns: List[str], scaler=StandardScaler(),
                   inplace: bool = False) -> pd.DataFrame:
    """
    Apply sklearn scaler to columns.

    :param df: input dataframe
    :param columns: List of columns to encode
    :param scaler: scaler object from sklearn
    :param inplace: If False, return a copy. Otherwise, do operation inplace and return None
    :return:
        - df:
          DataFrame scaled
        - dict_cls_fitted:
          dict by col of fitted cls
    """

    check_exist_vars(df, columns)

    if not inplace:
        df = df.copy(deep=True)

    le = scaler
    dict_cls_fitted = {}
    for col in columns:
        try:
            df[col] = le.fit_transform(df[col])
        except ValueError:
            df[col] = le.fit_transform(pd.DataFrame(df[col]))
        except TypeError:
            # TypeError: '<' not supported between instances of 'str' and 'float'
            # allows to encode nan
            df[col] = le.fit_transform(df[col].to_list())
        dict_cls_fitted[col] = le
    return df, dict_cls_fitted


def tsne(df: pd.DataFrame, columns: List[str], n_components: int = 2, random_state: Union[int, None] = 123,
         n_jobs: Union[int, None] = -1):
    """
    t-distributed Stochastic Neighbor Embedding.

    t-SNE is a tool to visualize high-dimensional data. It converts
    similarities between data points to joint probabilities and tries
    to minimize the Kullback-Leibler divergence between the joint
    probabilities of the low-dimensional embedding and the
    high-dimensional data. t-SNE has a cost function that is not convex,
    i.e. with different initializations we can get different results.

    :param df: input dataframe
    :param columns: List of columns to use
    :param n_components: Dimension of the embedded space. Default 2.
    :param random_state: int, RandomState instance or None, optional, default: 123
        If int, random_state is the seed used by the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.
    :param n_jobs: default=-1
        The number of parallel jobs to run for neighbors search. This parameter
        has no impact when ``metric="precomputed"`` or
        (``metric="euclidean"`` and ``method="exact"``).
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
    :return: array-like with projections
    """
    tsne = TSNE(n_components=n_components, random_state=random_state, n_jobs=n_jobs)
    projections = tsne.fit_transform(df[columns])
    return projections


def u_map(df: pd.DataFrame, columns: List[str], n_components: int = 2, random_state: Union[int, None] = 123,
          n_jobs: Union[int, None] = -1):
    """
    Just like t-SNE, UMAP is a dimensionality reduction specifically designed for visualizing complex data in
    low dimensions (2D or 3D). As the number of data points increase, UMAP becomes more time efficient compared to TSNE.

    :param df: input dataframe
    :param columns: List of columns to use
    :param n_components: Dimension of the embedded space. Default 2.
    :param random_state: int, RandomState instance or None, optional, default: 123
        If int, random_state is the seed used by the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.
    :param n_jobs: default=-1
        The number of parallel jobs to run for neighbors search. This parameter
        has no impact when ``metric="precomputed"`` or
        (``metric="euclidean"`` and ``method="exact"``).
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
    :return: array-like with projections
    """
    _umap = umap.UMAP(n_components=n_components, init='random', random_state=random_state, n_jobs=n_jobs)
    projections = _umap.fit_transform(df[columns])
    return projections


def split_filename(file_name: str) -> Tuple[str, str, str]:
    """
    Split file_name into folder path name, basename, and extension name.

    :param file_name: full path
    :return: path name, basename, extension name
    """
    pth = os.path.dirname(file_name)
    f_name = os.path.basename(file_name)

    ext = None
    for special_ext in ['.nii.gz']:
        ext_len = len(special_ext)
        if f_name[-ext_len:].lower() == special_ext:
            ext = f_name[-ext_len:]
            f_name = f_name[:-ext_len] if len(f_name) > ext_len else ''
            break
    if not ext:
        f_name, ext = os.path.splitext(f_name)
    return pth, f_name, ext


def check_is_nii_exist(input_file_path: str) -> str:
    """
    Check if a directory exist.

    :param input_file_path: string of the path of the nii or nii.gz.
    :return: string if exist, else raise Error.
    :raise: FileNotFoundError or FileExistsError
    """
    if not os.path.exists(input_file_path):
        raise FileNotFoundError(f"{input_file_path} was not found, check if it's a valid file path")

    pth, fnm, ext = split_filename(input_file_path)
    if ext not in [".nii", ".nii.gz"]:
        raise FileExistsError(f"extension of {input_file_path} need to be '.nii' or '.nii.gz'")
    return input_file_path


def load_nifty_volume_as_array(input_path_file: str) -> Tuple[np.ndarray, Tuple[Tuple, Tuple, Tuple]]:
    """
    Load nifty image into numpy array [z,y,x] axis order.
    The output array shape is like [Depth, Height, Width].

    :param input_path_file: input path file, should be '*.nii' or '*.nii.gz'
    :return: a numpy data array, (with header)
    """
    img = sitk.ReadImage(input_path_file)
    data = sitk.GetArrayFromImage(img)

    origin, spacing, direction = img.GetOrigin(), img.GetSpacing(), img.GetDirection()
    return data, (origin, spacing, direction)


def save_to_nii(im: np.ndarray, header: (tuple, tuple, tuple), output_dir: str, filename: str, mode: str = "image",
                gzip: bool = True) -> None:
    """
    Save numpy array to nii.gz format to submit.

    :param im: array numpy
    :param header: header metadata (origin, spacing, direction).
    :param output_dir: Output directory.
    :param filename: Filename of the output file.
    :param mode: save as 'image' or 'label'
    :param gzip: zip nii (ie, nii.gz)
    """
    origin, spacing, direction = header
    if mode == "label":
        img = sitk.GetImageFromArray(im.astype(np.uint8))
    else:
        img = sitk.GetImageFromArray(im.astype(np.float32))
    img.SetOrigin(origin), img.SetSpacing(spacing), img.SetDirection(direction)

    if gzip:
        ext = ".nii.gz"
    else:
        ext = ".nii"

    if not os.path.exists(output_dir):
        os.makedirs(output_dir, exist_ok=True)
    sitk.WriteImage(img, os.path.join(output_dir, filename) + ext)


def mat_to_bytes(nrows: int, ncols: int, dtype: int = 32, out: str = "GB") -> float:
    """
    Calculate the size of a numpy array in bytes.

    .. note::
        code from: `<https://gist.github.com/dimalik/f4609661fb83e3b5d22e7550c1776b90>`_

    :param nrows: the number of rows of the matrix.
    :param ncols: the number of columns of the matrix.
    :param dtype: the size of each element in the matrix. Defaults to 32bits.
    :param out: the output unit. Defaults to gigabytes (GB)
    :returns: the size of the matrix in the given unit
    """
    sizes = {v: i for i, v in enumerate("BYTES KB MB GB TB".split())}
    return nrows * ncols * dtype / 8 / 1024. ** sizes[out]



import torch
import torch.nn as nn
import torch.nn.functional as F
import math

class BrainAgeLoss(nn.Module):
    """
    Combined loss function for brain age prediction tasks.
    
    This loss combines:
    1. Mean Absolute Error (MAE) for accurate age prediction
    2. Correlation penalty to ensure predictions correlate with true ages
    3. Age-specific weighting to address age bias
    4. Regularization to prevent overfitting to the mean
    
    Parameters:
    -----------
    alpha : float, default=1.0
        Weight for the correlation component.
    beta : float, default=0.3
        Weight for the bias regularization component.
    gamma : float, default=0.2
        Weight for the age-specific weighting component.
    eps : float, default=1e-8
        Small constant for numerical stability.
    smoothing : float, default=0.1
        Amount of label smoothing to apply.
    use_huber : bool, default=False
        Whether to use Huber loss instead of MAE for robustness.
    delta : float, default=1.0
        Parameter for Huber loss if used.
    """
    def __init__(self, alpha=1.0, beta=0.3, gamma=0.2, eps=1e-8, 
                 smoothing=0.1, use_huber=True, delta=1.0):
        super().__init__()
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.eps = eps
        self.smoothing = smoothing
        self.use_huber = use_huber
        self.delta = delta
        
    def forward(self, pred, target):
        """
        Calculate the combined loss.
        
        Parameters:
        -----------
        pred : torch.Tensor
            Model predictions, shape (batch_size,)
        target : torch.Tensor
            Ground truth ages, shape (batch_size,)
        demographic_factors : torch.Tensor, optional
            Demographic factors like sex, ethnicity that might influence the loss weighting,
            shape (batch_size, num_factors)
            
        Returns:
        --------
        loss : torch.Tensor
            The combined loss value
        metrics : dict
            Dictionary containing individual loss components for monitoring
        """
        # Apply label smoothing if enabled
        if self.smoothing > 0:
            # Calculate mean age in the batch
            mean_age = torch.mean(target)
            # Smooth targets toward the mean age
            target_smooth = target * (1 - self.smoothing) + mean_age * self.smoothing
        else:
            target_smooth = target
            
        # 1. Basic error term (MAE or Huber)
        if self.use_huber:
            error_term = F.huber_loss(pred, target_smooth, delta=self.delta, reduction='none')
        else:
            error_term = torch.abs(pred - target_smooth)
            
        # 2. Age-specific weighting (optional)
        if self.gamma > 0:
            # compute weights that give more importance to under-represented age groups
            age_weights = 1.0 + self.gamma * (target / torch.max(target))
            weighted_error = error_term * age_weights
            basic_loss = torch.mean(weighted_error)
        else:
            basic_loss = torch.mean(error_term)
            
        # 3. Correlation term
        # Center the data
        pred_centered = pred - torch.mean(pred)
        target_centered = target - torch.mean(target)
        
        # Calculate Pearson correlation
        pred_var = torch.sum(pred_centered ** 2)
        target_var = torch.sum(target_centered ** 2)
        correlation = torch.sum(pred_centered * target_centered) / (torch.sqrt(pred_var * target_var) + self.eps)
        
        # convert to a loss (1 - correlation)
        correlation_loss = 1.0 - correlation
        # 4. bias regularization to prevent prediction toward mean
        pred_std = torch.std(pred)
        target_std = torch.std(target)
        std_ratio = torch.abs(pred_std / (target_std + self.eps) - 1.0)
        
        # 5. Combine all terms
        combined_loss = basic_loss + self.alpha * correlation_loss + self.beta * std_ratio
        
        # Return individual components for monitoring
        metrics = {
            'mae': torch.mean(torch.abs(pred - target)),  # Pure MAE for monitoring
            'corr_loss': correlation_loss.item(),
            'std_ratio_loss': std_ratio.item(),
            'correlation': correlation.item(),
            'pred_std': pred_std.item(),
            'target_std': target_std.item()
        }
        
        return combined_loss, metrics
    
    def get_name(self):
        """Returns a descriptive name of the loss function and its parameters."""
        name = f"BrainAgeLoss_a{self.alpha}_b{self.beta}_g{self.gamma}"
        if self.use_huber:
            name += f"_huber{self.delta}"
        if self.smoothing > 0:
            name += f"_smooth{self.smoothing}"
        return name


import os
import pytest
from cvasl.harmonizers import (
    NeuroHarmonize, 
    Covbat, 
    NeuroCombat, 
    NestedComBat, 
    ComscanNeuroCombat,
    AutoCombat,
    RELIEF,
    CombatPlusPlus
)
from cvasl.dataset import MRIdataset, encode_cat_features


def load_datasets(shared_datadir):
    """Load test datasets and preprocess them."""
    input_paths = [
        os.path.realpath(shared_datadir / "TestingData_Site1_fake.csv"),
        os.path.realpath(shared_datadir / "TestingData_Site2_fake.csv"),
        os.path.realpath(shared_datadir / "TrainingData_Site1_fake.csv")
    ]
    # Using duplicate site_id (1, 2, 1) to test that harmonizers handle multiple datasets from the same site correctly
    input_sites = [1, 2, 1]

    mri_datasets = [
        MRIdataset(input_path, input_site, "participant_id", features_to_drop=[])
        for input_site, input_path in zip(input_sites, input_paths)
    ]
    
    for mri_dataset in mri_datasets:
        mri_dataset.preprocess()
    
    features_to_map = ['sex']
    encode_cat_features(mri_datasets, features_to_map)
    
    return mri_datasets


def test_neurocombat(shared_datadir):
    """Test whether the NeuroCombat harmonizer runs."""
    datasets = load_datasets(shared_datadir)
    
    features_to_harmonize = [
        'ACA_B_CoV', 'MCA_B_CoV', 'PCA_B_CoV', 'TotalGM_B_CoV',
        'ACA_B_CBF', 'MCA_B_CBF', 'PCA_B_CBF', 'TotalGM_B_CBF'
    ]
    discrete_covariates = ['sex']
    continuous_covariates = ['age']
    patient_identifier = 'participant_id'
    site_indicator = 'site'

    harmonizer = NeuroCombat(
        features_to_harmonize, 
        discrete_covariates, 
        continuous_covariates, 
        patient_identifier, 
        site_indicator
    )
    harmonized_data = harmonizer.harmonize(datasets)
    
    # Verify harmonization succeeded
    assert harmonized_data is not None
    assert len(harmonized_data) == len(datasets)
    
    # Verify data structure is preserved
    for original, harmonized in zip(datasets, harmonized_data):
        assert harmonized.data is not None
        assert len(harmonized.data.columns) > 0


def test_neuroharmonize(shared_datadir):
    """Test whether the NeuroHarmonize harmonizer runs."""
    datasets = load_datasets(shared_datadir)
    
    features_to_harmonize = [
        'aca_b_cov', 'mca_b_cov', 'pca_b_cov', 'totalgm_b_cov',
        'aca_b_cbf', 'mca_b_cbf', 'pca_b_cbf', 'totalgm_b_cbf'
    ]
    covariates = ['age', 'sex', 'site']
    site_indicator = 'site'
    
    harmonizer = NeuroHarmonize(
        features_to_harmonize=features_to_harmonize,
        covariates=covariates,
        smooth_terms=[],
        site_indicator=site_indicator,
        empirical_bayes=True
    )
    harmonized_data = harmonizer.harmonize(datasets)
    
    assert harmonized_data is not None
    assert len(harmonized_data) == len(datasets)
    
    # Check that harmonized features exist in output
    for dataset in harmonized_data:
        for feature in features_to_harmonize:
            assert feature in dataset.data.columns


def test_covbat(shared_datadir):
    """Test whether the CovBat harmonizer runs."""
    datasets = load_datasets(shared_datadir)
    
    patient_identifier = 'participant_id'
    site_indicator = 'site'
    features_to_harmonize = [
        'aca_b_cov', 'mca_b_cov', 'pca_b_cov', 'totalgm_b_cov',
        'aca_b_cbf', 'mca_b_cbf', 'pca_b_cbf', 'totalgm_b_cbf'
    ]
    numerical_covariates = ['age']
    covariates = ['age', 'sex']
    
    harmonizer = Covbat(
        features_to_harmonize=features_to_harmonize,
        covariates=covariates,
        site_indicator=site_indicator,
        patient_identifier=patient_identifier,
        numerical_covariates=numerical_covariates
    )
    harmonized_data = harmonizer.harmonize(datasets)
    
    assert harmonized_data is not None
    assert len(harmonized_data) == len(datasets)


@pytest.mark.skip(reason="Requires multiple samples per covariate value (test data too small)")
def test_comscan_neurocombat(shared_datadir):
    """Test whether the ComscanNeuroCombat harmonizer runs."""
    datasets = load_datasets(shared_datadir)
    
    features_to_harmonize = [
        'aca_b_cov', 'mca_b_cov', 'pca_b_cov', 'totalgm_b_cov',
        'aca_b_cbf', 'mca_b_cbf', 'pca_b_cbf', 'totalgm_b_cbf'
    ]
    discrete_covariates = ['sex']
    continuous_covariates = ['age']
    site_indicator = 'site'
    
    harmonizer = ComscanNeuroCombat(
        features_to_harmonize=features_to_harmonize,
        discrete_covariates=discrete_covariates,
        continuous_covariates=continuous_covariates,
        site_indicator=site_indicator
    )
    harmonized_data = harmonizer.harmonize(datasets)
    
    # ComscanNeuroCombat may return None if no features to harmonize
    if harmonized_data is not None:
        assert len(harmonized_data) == len(datasets)


@pytest.mark.skip(reason="Not enough samples for clustering algorithm (test data too small)")
def test_autocombat(shared_datadir):
    """Test whether the AutoCombat harmonizer runs."""
    datasets = load_datasets(shared_datadir)
    
    features_to_harmonize = [
        'aca_b_cov', 'mca_b_cov', 'pca_b_cov', 'totalgm_b_cov',
        'aca_b_cbf', 'mca_b_cbf', 'pca_b_cbf', 'totalgm_b_cbf'
    ]
    data_subset = [
        'aca_b_cov', 'mca_b_cov', 'pca_b_cov', 'totalgm_b_cov',
        'aca_b_cbf', 'mca_b_cbf', 'pca_b_cbf', 'totalgm_b_cbf',
        'site', 'sex', 'age'
    ]
    discrete_covariates = ['sex']
    continuous_covariates = ['age']
    site_indicator = ['site']
    
    harmonizer = AutoCombat(
        data_subset=data_subset,
        features_to_harmonize=features_to_harmonize,
        site_indicator=site_indicator,
        discrete_covariates=discrete_covariates,
        continuous_covariates=continuous_covariates
    )
    harmonized_data = harmonizer.harmonize(datasets)
    
    if harmonized_data is not None:
        assert len(harmonized_data) == len(datasets)


@pytest.mark.skipif(
    not hasattr(pytest, 'R_AVAILABLE') or not getattr(pytest, 'R_AVAILABLE', False),
    reason="R dependencies (rpy2) not available"
)
@pytest.mark.skip(reason="Requires R packages (denoiseR, RcppCNPy) which may not be available in all environments")
def test_relief(shared_datadir):
    """Test whether the RELIEF harmonizer runs (requires R)."""
    datasets = load_datasets(shared_datadir)
    
    features_to_harmonize = [
        'aca_b_cov', 'mca_b_cov', 'pca_b_cov', 'totalgm_b_cov',
        'aca_b_cbf', 'mca_b_cbf', 'pca_b_cbf', 'totalgm_b_cbf'
    ]
    covariates = ['sex', 'age']
    patient_identifier = 'participant_id'
    
    harmonizer = RELIEF(
        features_to_harmonize=features_to_harmonize,
        covariates=covariates,
        patient_identifier=patient_identifier
    )
    harmonized_data = harmonizer.harmonize(datasets)
    
    assert harmonized_data is not None
    assert len(harmonized_data) == len(datasets)


@pytest.mark.skip(reason="File path issue trips up test on Windows")
@pytest.mark.skipif(
    not hasattr(pytest, 'R_AVAILABLE') or not getattr(pytest, 'R_AVAILABLE', False),
    reason="R dependencies (rpy2) not available"
)
def test_combat_plusplus(shared_datadir):
    """Test whether the CombatPlusPlus harmonizer runs (requires R)."""
    datasets = load_datasets(shared_datadir)
    
    features_to_harmonize = [
        'aca_b_cov', 'mca_b_cov', 'pca_b_cov', 'totalgm_b_cov',
        'aca_b_cbf', 'mca_b_cbf', 'pca_b_cbf', 'totalgm_b_cbf'
    ]
    discrete_covariates = ['sex']
    continuous_covariates = ['age']
    discrete_covariates_to_remove = []
    continuous_covariates_to_remove = []
    site_indicator = 'site'
    patient_identifier = 'participant_id'
    
    harmonizer = CombatPlusPlus(
        features_to_harmonize=features_to_harmonize,
        discrete_covariates=discrete_covariates,
        continuous_covariates=continuous_covariates,
        discrete_covariates_to_remove=discrete_covariates_to_remove,
        continuous_covariates_to_remove=continuous_covariates_to_remove,
        site_indicator=site_indicator,
        patient_identifier=patient_identifier
    )
    harmonized_data = harmonizer.harmonize(datasets)
    
    assert harmonized_data is not None
    assert len(harmonized_data) == len(datasets)


def test_nested_combat(shared_datadir):
    """Test whether the NestedComBat harmonizer runs."""
    datasets = load_datasets(shared_datadir)
    
    features_to_harmonize = [
        'ACA_B_CoV', 'MCA_B_CoV', 'PCA_B_CoV', 'TotalGM_B_CoV',
        'ACA_B_CBF', 'MCA_B_CBF', 'PCA_B_CBF', 'TotalGM_B_CBF'
    ]
    batch_list_harmonisations = ['site']  # Simplified for test data
    site_indicator = ['site']
    discrete_covariates = ['sex']
    continuous_covariates = ['age']
    patient_identifier = 'participant_id'
    
    harmonizer = NestedComBat(
        features_to_harmonize=features_to_harmonize,
        batch_list_harmonisations=batch_list_harmonisations,
        site_indicator=site_indicator,
        discrete_covariates=discrete_covariates,
        continuous_covariates=continuous_covariates,
        patient_identifier=patient_identifier,
        intermediate_results_path='.',
        return_extended=False,
        use_gmm=False
    )
    harmonized_data = harmonizer.harmonize(datasets)
    
    assert harmonized_data is not None
    assert len(harmonized_data) == len(datasets)


# Test for error handling
def test_neurocombat_with_invalid_input(shared_datadir):
    """Test that NeuroCombat raises appropriate errors with invalid input."""
    datasets = load_datasets(shared_datadir)
    
    # Test with empty features list
    with pytest.raises((ValueError, TypeError)):
        harmonizer = NeuroCombat(
            features_to_harmonize=[],
            discrete_covariates=['sex'],
            continuous_covariates=['age'],
            patient_identifier='participant_id',
            site_indicator='site'
        )


def test_harmonizer_preserves_patient_ids(shared_datadir):
    """Test that harmonization preserves patient identifiers."""
    datasets = load_datasets(shared_datadir)
    
    # Store original patient IDs
    original_ids = [set(dataset.data['participant_id'].values) for dataset in datasets]
    
    features_to_harmonize = [
        'ACA_B_CoV', 'MCA_B_CoV', 'PCA_B_CoV', 'TotalGM_B_CoV',
        'ACA_B_CBF', 'MCA_B_CBF', 'PCA_B_CBF', 'TotalGM_B_CBF'
    ]
    
    harmonizer = NeuroCombat(
        features_to_harmonize=features_to_harmonize,
        discrete_covariates=['sex'],
        continuous_covariates=['age'],
        patient_identifier='participant_id',
        site_indicator='site'
    )
    harmonized_data = harmonizer.harmonize(datasets)
    
    # Check that patient IDs are preserved
    for original_id_set, harmonized_dataset in zip(original_ids, harmonized_data):
        harmonized_id_set = set(harmonized_dataset.data['participant_id'].values)
        assert original_id_set == harmonized_id_set, "Patient IDs should be preserved"


def test_harmonizer_output_shape(shared_datadir):
    """Test that harmonized data has the expected shape."""
    datasets = load_datasets(shared_datadir)
    
    # Store original shapes
    original_shapes = [(len(dataset.data), len(dataset.data.columns)) 
                       for dataset in datasets]
    
    features_to_harmonize = [
        'ACA_B_CoV', 'MCA_B_CoV', 'PCA_B_CoV', 'TotalGM_B_CoV',
        'ACA_B_CBF', 'MCA_B_CBF', 'PCA_B_CBF', 'TotalGM_B_CBF'
    ]
    
    harmonizer = NeuroCombat(
        features_to_harmonize=features_to_harmonize,
        discrete_covariates=['sex'],
        continuous_covariates=['age'],
        patient_identifier='participant_id',
        site_indicator='site'
    )
    harmonized_data = harmonizer.harmonize(datasets)
    
    # Check that shapes match (rows should be same, columns should be preserved)
    for (orig_rows, orig_cols), harmonized_dataset in zip(original_shapes, harmonized_data):
        harm_rows = len(harmonized_dataset.data)
        harm_cols = len(harmonized_dataset.data.columns)
        
        # Number of rows should be preserved
        assert harm_rows == orig_rows, f"Row count mismatch: {harm_rows} vs {orig_rows}"
        # Column count should be similar (allowing for minor differences in processing)
        assert abs(harm_cols - orig_cols) <= 2, f"Column count significantly different: {harm_cols} vs {orig_cols}"



import sys

import warnings

import pandas as pd
from scipy import stats


def encode_cat_features(dff,cat_features_to_encode):

    feature_mappings = {}
    reverse_mappings = {}
    
    # Track the original lengths to split data back correctly
    dataset_lengths = [len(_d.data) for _d in dff]
    
    data = pd.concat([_d.data for _d in dff], ignore_index=True)

    for feature in cat_features_to_encode:
        if feature in data.columns:
            unique_values = data[feature].unique()
            mapping = {value: i for i, value in enumerate(unique_values)}
            feature_mappings[feature] = mapping
            reverse_mappings[feature] = {v: k for k, v in mapping.items()}
            data[feature] = data[feature].map(mapping)

    # Split data back using original dataset lengths
    start_idx = 0
    for i, _d in enumerate(dff):
        end_idx = start_idx + dataset_lengths[i]
        _d.data = data.iloc[start_idx:end_idx].reset_index(drop=True)
        _d.feature_mappings = feature_mappings
        _d.reverse_mappings = reverse_mappings
        _d.cat_features_to_encode = cat_features_to_encode
        start_idx = end_idx
    return dff

class MRIdataset:
    def __init__(
        self,
        path,
        site_id,
        patient_identifier="participant_id",
        cat_features_to_encode=None,
        ICV=False,
        decade=False,
        features_to_drop=[],
        features_to_bin=None,
        binning_method="equal_width",
        num_bins=10,
        bin_labels=None,
    ):
        """
        Initializes the MRIdataset class.

        Args:
            path (str or list): Path to the CSV file or list of paths.
            site_id (int or str): Identifier for the site.
            patient_identifier (str, optional): Column name for patient ID. Defaults to "participant_id".
            cat_features_to_encode (list, optional): List of categorical features to encode. Defaults to None.
            ICV (bool, optional): Whether to add ICV-related features. Defaults to False.
            decade (bool, optional): Whether to add decade-related features. Defaults to False.
            features_to_drop (list, optional): List of features to drop. Defaults to ["m0", "id"].
            features_to_bin (list, optional): List of features to bin. Defaults to None.
            binning_method (str, optional): Binning method to use ("equal_width", "equal_frequency"). Defaults to "equal_width".
            num_bins (int, optional): Number of bins. Defaults to 10.
            bin_labels (list, optional): Labels for bins. Defaults to None.
        """
        self._validate_init_arguments(
            path,
            site_id,
            patient_identifier,
            cat_features_to_encode,
            ICV,
            decade,
            features_to_drop,
            features_to_bin,
            binning_method,
            num_bins,
            bin_labels,
        )

        if isinstance(path, list):
            self.data = pd.concat([pd.read_csv(_p) for _p in path])
        else:
            self.data = pd.read_csv(path)

        self.site_id = site_id
        self.data["Site"] = self.site_id
        self.feature_mappings = {}
        self.reverse_mappings = {}
        self.patient_identifier = patient_identifier
        self.icv = ICV
        self.path = path
        self.decade = decade
        self.features_to_drop = features_to_drop
        self.fetures_to_bin = features_to_bin
        self.binning_method = binning_method
        self.num_bins = num_bins
        self.bin_labels = bin_labels
        self.cat_features_to_encode = cat_features_to_encode
        self.initial_statistics = None
        self.harmonized_statistics = None
        self.columns_order = self.data.columns.to_list()
        self.dropped_features = None

    def _validate_init_arguments(
        self,
        path,
        site_id,
        patient_identifier,
        cat_features_to_encode,
        ICV,
        decade,
        features_to_drop,
        features_to_bin,
        binning_method,
        num_bins,
        bin_labels,
    ):
        """Validates arguments passed to the __init__ method."""
        if not isinstance(path, (str, list)):
            raise TypeError("path must be a string or a list of strings.")
        if isinstance(path, list) and not all(isinstance(p, str) for p in path):
            raise TypeError("If path is a list, all elements must be strings.")
        if not isinstance(site_id, (int, str)):
            raise TypeError("site_id must be an integer or a string.")
        if not isinstance(patient_identifier, str):
            raise TypeError("patient_identifier must be a string.")
        if cat_features_to_encode is not None and not isinstance(cat_features_to_encode, list):
            raise TypeError("cat_features_to_encode must be a list or None.")
        if not isinstance(ICV, bool):
            raise TypeError("ICV must be a boolean.")
        if not isinstance(decade, bool):
            raise TypeError("decade must be a boolean.")
        if not isinstance(features_to_drop, list):
            raise TypeError("features_to_drop must be a list.")
        if features_to_bin is not None and not isinstance(features_to_bin, list):
            raise TypeError("features_to_bin must be a list or None.")
        if not isinstance(binning_method, str):
            raise TypeError("binning_method must be a string.")
        if binning_method not in ["equal_width", "equal_frequency"]:
            raise ValueError("binning_method must be either 'equal_width' or 'equal_frequency'.")
        if not isinstance(num_bins, int) or num_bins <= 0:
            raise ValueError("num_bins must be a positive integer.")
        if bin_labels is not None and not isinstance(bin_labels, list):
            raise TypeError("bin_labels must be a list or None.")

    def generalized_binning(self):
        """
        Performs binning on specified features of a Pandas DataFrame.

        Args:
            df: The input DataFrame.
            features: A list of feature names (columns) to bin.
            binning_method: The binning method to use. Options are 'equal_width', 'equal_frequency'.
                Defaults to 'equal_width'.
            num_bins: The number of bins to create (for 'equal_width' and 'equal_frequency'). Defaults to 10.
            labels: Optional labels for the resulting bins (for 'equal_width' and 'equal_frequency').

        Returns:
            A new DataFrame with the binned features.  Returns None if an invalid binning method is specified.
        """

        features = self.fetures_to_bin
        binning_method = self.binning_method
        num_bins = self.num_bins
        labels = self.bin_labels

        df_binned = (
            self.data.copy()
        )  # Create a copy to avoid modifying the original DataFrame

        for feature in features:
            if binning_method == "equal_width":
                df_binned[feature + "_binned"], bins = pd.cut(
                    self.data[feature],
                    bins=num_bins,
                    labels=labels,
                    retbins=True,
                    duplicates="drop",
                )
            elif binning_method == "equal_frequency":
                df_binned[feature + "_binned"], bins = pd.qcut(
                    self.data[feature],
                    q=num_bins,
                    labels=labels,
                    retbins=True,
                    duplicates="drop",
                )
            else:
                return None  # Handle invalid binning methods

        self.data = df_binned

    def encode_categorical_features(self):

        for feature in self.cat_features_to_encode:
            if feature in self.data.columns:
                unique_values = self.data[feature].unique()
                mapping = {value: i for i, value in enumerate(unique_values)}
                self.feature_mappings[feature] = mapping
                self.reverse_mappings[feature] = {v: k for k, v in mapping.items()}
                self.data[feature] = self.data[feature].map(mapping)

    def reverse_encode_categorical_features(self):
        for feature, mapping in self.reverse_mappings.items():
            if feature in self.data.columns:
                unique_values = self.data[feature].unique()
                # Filter the mapping to include only the unique values
                filtered_mapping = {k: mapping[k] for k in unique_values if k in mapping}
                self.data[feature] = self.data[feature].map(filtered_mapping)

    def addICVfeatures(self):
        self.data["icv"] = self.data["gm_vol"] / self.data["gm_icvratio"]

    def addDecadefeatures(self):
        self.data["decade"] = (self.data["age"] / 10).round()
        #self.data = self.data.sort_values(by="age")
        self.data.reset_index(inplace=True)

    def dropFeatures(self):

        self.dropped_features = self.data[[self.patient_identifier] + self.features_to_drop]
        self.data = self.data.drop(self.features_to_drop, axis=1)

    #define a function that would add site_id to the bginning of the patient_identifier column of the data
    def unique_patient_ids(self):
        self.data[self.patient_identifier] = "MRIPatient_Site:" + str(self.site_id) + '_OriginalID:'  + self.data[self.patient_identifier]

    #define a function that restores the patient_identifier column to its original state
    def restore_patient_ids(self):
        self.data[self.patient_identifier] = self.data[self.patient_identifier].str.split('_OriginalID:').str[1]

    def prepare_for_export(self):
        self.restore_patient_ids()
        if self.dropped_features is not None:
            self.data = self.data.merge(self.dropped_features,on=self.patient_identifier)
        for _c in ['index','Index' ,'ID','unnamed: 0']:
            if _c in self.data.columns:
                self.data = self.data.drop(columns=[_c])
        self.reverse_encode_categorical_features()
        _tc = [_c.lower() for _c in self.columns_order]
        self.data = self.data[_tc]
        self.data.columns = self.columns_order
        for _c in ['index','Index' ,'ID','unnamed: 0']:
            if _c in self.data.columns:
                self.data = self.data.drop(columns=[_c])


    def _extended_summary_statistics(self):
        """
        Calculates extended summary statistics for each column of a Pandas DataFrame.

        Args:
            df: The input DataFrame.

        Returns:
            A Pandas DataFrame containing the summary statistics.
        """
        df = self.data
        summary_stats = []

        for col_name in df.columns:
            col = df[col_name]
            col_type = col.dtype

            stats_dict = {
                "Column Name": col_name,
                "Data Type": col_type,
                "Count": col.count(),
                "Number of Unique Values": col.nunique(),
                "Missing Values": col.isnull().sum(),
            }

            if pd.api.types.is_numeric_dtype(col_type):
                stats_dict.update({
                    "Mean": col.mean(),
                    "Standard Deviation": col.std(),
                    "Minimum": col.min(),
                    "25th Percentile": col.quantile(0.25),
                    "Median (50th Percentile)": col.median(),
                    "75th Percentile": col.quantile(0.75),
                    "Maximum": col.max(),
                    "Skewness": col.skew(),
                    "Kurtosis": col.kurt(),
                })
                if len(col.dropna()) > 1:  # Check for sufficient data points for normality test
                    statistic, p_value = stats.shapiro(col.dropna())
                    stats_dict.update({
                        "Shapiro-Wilk Test Statistic": statistic,
                        "Shapiro-Wilk p-value": p_value
                    })

            elif pd.api.types.is_categorical_dtype(col_type) or pd.api.types.is_object_dtype(col_type):
                mode = col.mode()
                mode_str = ', '.join(mode.astype(str)) # handles multiple modes
                stats_dict.update({
                    "Mode": mode_str
                })
                top_n = 5  # Display top N most frequent categories
                value_counts_df = col.value_counts().nlargest(top_n).reset_index()
                value_counts_df.columns = ['Value', 'Count']
                for i in range(len(value_counts_df)):
                    stats_dict[f"Top {i+1} Most Frequent Value"] = value_counts_df.iloc[i,0]
                    stats_dict[f"Top {i+1} Most Frequent Value Count"] = value_counts_df.iloc[i,1]

            summary_stats.append(stats_dict)

        return pd.DataFrame(summary_stats)


    def preprocess(self):
        """
        Applies preprocessing steps to the MRIdataset.

        This includes:
            - Lowercasing column names.
            - Dropping specified features.
            - Creating unique patient IDs.
            - Adding decade-based features (optional).
            - Adding ICV-related features (optional).
            - Encoding categorical features (optional).
            - Binning specified features (optional).
        """
        # Common preprocessing steps
        self.data.columns = self.data.columns.str.lower()
        if self.features_to_drop:
            self._drop_features() # Use private method

        self._unique_patient_ids() # Use private method

        if self.decade:
            self._add_decade_features() # Use private method
        if self.icv:
            self._add_icv_features() # Use private method
        if self.cat_features_to_encode:
            self._encode_categorical_features() # Use private method
        if self.fetures_to_bin:
            self._generalized_binning() # Use private method
        #self.initial_statistics = self._extended_summary_statistics() # keep for now, decide later if needed

    def _drop_features(self):
        """Drops specified features from the dataset."""
        self.dropped_features = self.data[[self.patient_identifier] + self.features_to_drop]
        self.data = self.data.drop(self.features_to_drop, axis=1, errors='ignore') # Added errors='ignore'

    def _unique_patient_ids(self):
        """Adds site_id to the beginning of the patient_identifier column to ensure uniqueness across sites."""
        self.data[self.patient_identifier] = "MRIPatient_Site:" + str(self.site_id) + '_OriginalID:'  + self.data[self.patient_identifier]

    def _add_decade_features(self):
        """Adds decade-based features based on the 'age' column."""
        if 'age' in self.data.columns:
            self.data["decade"] = (self.data["age"] / 10).round()
            self.data.reset_index(inplace=True, drop=True) # reset index after sorting and inplace=True
        else:
            warnings.warn("Age column not found, cannot add decade features.")

    def _add_icv_features(self):
        """Adds ICV-related features by calculating 'icv' ratio."""
        if 'gm_vol' in self.data.columns and 'gm_icvratio' in self.data.columns:
            self.data["icv"] = self.data["gm_vol"] / self.data["gm_icvratio"]
        else:
            warnings.warn("Required columns 'gm_vol' or 'gm_icvratio' not found, cannot add ICV features.")

    def _encode_categorical_features(self):
        """Encodes categorical features specified in self.cat_features_to_encode."""
        for feature in self.cat_features_to_encode:
            if feature in self.data.columns:
                unique_values = self.data[feature].unique()
                mapping = {value: i for i, value in enumerate(unique_values)}
                self.feature_mappings[feature] = mapping
                self.reverse_mappings[feature] = {v: k for k, v in mapping.items()}
                self.data[feature] = self.data[feature].map(mapping)
            else:
                warnings.warn(f"Categorical feature '{feature}' not found in dataset, skipping encoding.")

    def _generalized_binning(self):
        """Applies generalized binning to features specified in self.fetures_to_bin."""
        features = self.fetures_to_bin
        binning_method = self.binning_method
        num_bins = self.num_bins
        labels = self.bin_labels

        df_binned = self.data.copy() # Create copy to avoid modifying original

        for feature in features:
            if feature not in self.data.columns:
                warnings.warn(f"Feature '{feature}' not found in dataset, skipping binning.")
                continue # Skip to next feature if current one is not found

            if binning_method == "equal_width":
                df_binned[feature + "_binned"], bins = pd.cut(
                    self.data[feature],
                    bins=num_bins,
                    labels=labels,
                    retbins=True,
                    duplicates="drop",
                )
            elif binning_method == "equal_frequency":
                df_binned[feature + "_binned"], bins = pd.qcut(
                    self.data[feature],
                    q=num_bins,
                    labels=labels,
                    retbins=True,
                    duplicates="drop",
                )
            else:
                warnings.warn(f"Invalid binning method '{binning_method}', skipping binning for feature '{feature}'.")
                continue # Skip to next feature if binning method is invalid

        self.data = df_binned



import torch
import torch.nn as nn
from .resnext3d import SEBlock3D

class PreActResBlock3D(nn.Module):
    """Pre-activation ResNet3D Block with Bottleneck and SE."""

    def __init__(self, in_channels, out_channels, stride=1):
        super().__init__()
        bottleneck_channels = out_channels // 4  # bottleneck channels
        self.bn1 = nn.BatchNorm3d(in_channels)
        self.relu = nn.ReLU(inplace=True)
        self.conv1 = nn.Conv3d(in_channels, bottleneck_channels, kernel_size=1, bias=False)

        self.bn2 = nn.BatchNorm3d(bottleneck_channels)
        self.conv2 = nn.Conv3d(bottleneck_channels, bottleneck_channels, kernel_size=3, stride=stride, padding=1, bias=False)

        self.bn3 = nn.BatchNorm3d(bottleneck_channels)
        self.conv3 = nn.Conv3d(bottleneck_channels, out_channels, kernel_size=1, bias=False)

        self.se = SEBlock3D(out_channels)  # add squeeze-and-excitation

        self.shortcut = nn.Sequential()
        if stride != 1 or in_channels != out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv3d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False)
            )


    def forward(self, x):
        identity = x

        out = self.bn1(x)
        out = self.relu(out)
        out = self.conv1(out)

        out = self.bn2(out)
        out = self.relu(out)
        out = self.conv2(out)

        out = self.bn3(out)
        out = self.relu(out)
        out = self.conv3(out)

        out = self.se(out)

        out += self.shortcut(identity)
        return out



class Improved3DCNN(nn.Module):
    def __init__(self, num_demographics, initial_filters=32, filters_multiplier=2, num_conv_layers=3, use_se=True, dropout_rate=0.3, use_demographics=False):
        super().__init__()
        self.initial_filters = initial_filters
        self.filters_multiplier = filters_multiplier
        self.num_conv_layers = num_conv_layers
        self.use_se = use_se
        self.dropout_rate = dropout_rate
        self.use_demographics = use_demographics

        current_filters = initial_filters
        self.conv1 = nn.Conv3d(1, current_filters, kernel_size=3, padding=1)
        self.bn1 = nn.BatchNorm3d(current_filters)
        self.relu = nn.ReLU(inplace=True)
        self.pool1 = nn.MaxPool3d(2)

        # dynamically create convolutional layers
        conv_layers = []
        for i in range(num_conv_layers -1): # -1 because first conv is outside
            in_channels = current_filters
            current_filters = int(current_filters * filters_multiplier)
            out_channels = current_filters
                         
            conv_layer = nn.Conv3d(in_channels, out_channels, kernel_size=3, padding=1)
            conv_layers.append(conv_layer)
            conv_layers.append(nn.BatchNorm3d(out_channels))
            conv_layers.append(SEBlock3D(out_channels) if use_se else nn.Identity())
            conv_layers.append(nn.ReLU(inplace=True))
            conv_layers.append(nn.MaxPool3d(2))
            

        self.conv_layers = nn.Sequential(*conv_layers)


        self.global_avg_pool = nn.AdaptiveAvgPool3d(1)
        self.flatten = nn.Flatten()

        # correct flattened_size calculation, now it's truly dynamic
        flattened_size = current_filters  # after all multiplications


        self.fc1 = nn.Linear(flattened_size, 128)
        self.dropout = nn.Dropout(p=dropout_rate)
        fc2_input_size = 128 + num_demographics if self.use_demographics else 128
        self.fc2 = nn.Linear(fc2_input_size, 1)

            
    def forward(self, x, demographics):
        x = self.pool1(self.relu(self.bn1(self.conv1(x))))
        x = self.conv_layers(x)  # apply the dynamic conv layers
        x = self.global_avg_pool(x)
        x = self.flatten(x)
        x = self.dropout(self.relu(self.fc1(x)))
        if self.use_demographics:
            x = torch.cat((x, demographics), dim=1)
        x = self.fc2(x)
        return x

    def get_name(self):
        name = f"Improved3DCNN_layers{self.num_conv_layers}_filters{self.initial_filters}_filtermult{self.filters_multiplier}"
        if self.use_se:
            name += "_SE"
        name += f"_dropout{self.dropout_rate}"
        if self.use_demographics:
            name += "_with_demographics"  # indicate when using demographics
        else:
            name += "_without_demographics"
        return name

    def get_params(self):
        return {
            "initial_filters": self.initial_filters,
            "filters_multiplier": self.filters_multiplier,
            "num_conv_layers": self.num_conv_layers,
            "use_se": self.use_se,
            "dropout_rate": self.dropout_rate,
            "use_demographics": self.use_demographics,
            "architecture": "Improved3DCNN",
        }




# -*- coding: utf-8 -*-
"""
| Author: Alexandre CARRE 
| Created on: Jan 14, 2021
"""
import time
import warnings
from collections.abc import Iterable
from typing import Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import umap
from joblib import Parallel, delayed
from kneed import KneeLocator
from scipy.spatial.distance import cdist
from sklearn.base import TransformerMixin, ClusterMixin, BaseEstimator
from sklearn.base import clone
from sklearn.decomposition import PCA
from yellowbrick.cluster.elbow import ClusteringScoreVisualizer, KELBOW_SCOREMAP, YellowbrickValueError, \
    YellowbrickWarning
from yellowbrick.style.palettes import LINE_COLOR

with warnings.catch_warnings():
    # Ignore flood of RuntimeWarning: Explicit initial center position passed: performing only one init
    # in k-means instead of n_init=10 return_n_iter=True)
    warnings.filterwarnings("ignore", category=RuntimeWarning)
    from k_means_constrained import KMeansConstrained


class KMeansConstrainedMissing(TransformerMixin, ClusterMixin, BaseEstimator):
    """
    K-Means clustering with minimum and maximum cluster size constraints with possible missing values

    .. note::
        inspired of `<https://stackoverflow.com/questions/35611465/python-scikit-learn-clustering-with-missing-data>`_

    Parameters
    ----------
    n_clusters : int, optional, default: 8
        The number of clusters to form as well as the number of
        centroids to generate.

    size_min : int, optional, default: None
        Constrain the label assignment so that each cluster has a minimum
        size of size_min. If None, no constrains will be applied

    size_max : int, optional, default: None
        Constrain the label assignment so that each cluster has a maximum
        size of size_max. If None, no constrains will be applied

    em_iter :  int, default: 10
        expectation–maximization (EM) iteration for convergence of missing
        values. Use when no features reduction is applied and missing values.

    n_init : int, default: 10
        Number of times the k-means algorithm will be run with different
        centroid seeds. The final results will be the best output of
        n_init consecutive runs in terms of inertia.

    max_iter : int, default: 300
        Maximum number of iterations of the k-means algorithm for a
        single run.

    features_reduction : str, default: None
        Method for reduction of the embedded space with n_components. Can be pca or umap.

    n_components : int, default: 2
         Dimension of the embedded space for features reduction.

    tol : float, default: 1e-4
        Relative tolerance with regards to inertia to declare convergence

    verbose : int, default: 0
        Verbosity mode.

    random_state : int, RandomState instance or None, optional, default: None
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    copy_x : boolean, default: True
        When pre-computing distances it is more numerically accurate to center
        the data first.  If copy_x is True, then the original data is not
        modified.  If False, the original data is modified, and put back before
        the function returns, but small numerical differences may be introduced
        by subtracting and then adding the data mean.

    n_jobs : int, default: 1
        The number of jobs to use for the computation. This works by computing
        each of the n_init runs in parallel.

        If -1 all CPUs are used. If 1 is given, no parallel computing code is
        used at all, which is useful for debugging. For n_jobs below -1,
        (n_cpus + 1 + n_jobs) are used. Thus for n_jobs = -2, all CPUs but one
        are used.

    Attributes
    ----------
    cluster_centers\_ : array, [n_clusters, n_features]
        Coordinates of cluster centers

    labels\_ :
        Labels of each point

    inertia\_ : float
        Sum of squared distances of samples to their closest cluster center.

    cls\_ : KMeansConstrained classifier object

    cls_features_reduction\_ : PCA or UMAP reduction object

    centroids\_: array
        Centroids found at the last iteration of k-means.

    X_hat\_ : array
        Copy of X with the missing values filled in.

    mu\_ : Columns means

    Examples
    --------
    >>> X = np.array([[1, 2], [1, 4], [1, 0],
    ...                [4, 2], [4, 4], [4, 0]])
    >>> clf = KMeansConstrainedMissing(
    ...     n_clusters=2,
    ...     size_min=2,
    ...     size_max=5,
    ...     random_state=0
    ... )
    >>> clf.fit_predict(X)
    array([0, 0, 0, 1, 1, 1], dtype=int32)
    >>> clf.cluster_centers_
    array([[ 1.,  2.],
           [ 4.,  2.]])
    >>> clf.labels_
    array([0, 0, 0, 1, 1, 1], dtype=int32)

    Notes
    ------
    K-means problem constrained with a minimum and/or maximum size for each cluster.

    The constrained assignment is formulated as a Minimum Cost Flow (MCF) linear network optimisation
    problem. This is then solved using a cost-scaling push-relabel algorithm. The implementation used is
    Google's Operations Research tools's `SimpleMinCostFlow`.

    Ref:
        1. Bradley, P. S., K. P. Bennett, and Ayhan Demiriz. "Constrained k-means clustering."
        Microsoft Research, Redmond (2000): 1-8.
        2. Google's SimpleMinCostFlow implementation:
        https://github.com/google/or-tools/blob/master/ortools/graph/min_cost_flow.h
    """

    def __init__(self, n_clusters=8,
                 size_min=None,
                 size_max=None,
                 em_iter=10,
                 n_init=10,
                 max_iter=300,
                 features_reduction: Optional[str] = None,
                 n_components: int = 2,
                 tol=1e-4,
                 verbose=False,
                 random_state=None,
                 copy_x=True,
                 n_jobs=1):

        self.n_clusters = n_clusters
        self.size_min = size_min
        self.size_max = size_max
        self.em_iter = em_iter
        self.n_init = n_init
        self.max_iter = max_iter
        self.features_reduction = features_reduction
        self.n_components = n_components
        self.tol = tol
        self.verbose = verbose
        self.random_state = random_state
        self.copy_x = copy_x
        self.n_jobs = n_jobs

    def fit(self, X, y=None):
        """Compute k-means clustering with given constants.

        Parameters
        ----------
        X : array-like, shape=(n_samples, n_features)
            Training instances to cluster.

        y : Ignored

        """
        #################################################
        # for i in range(X.shape[1]):
        #     print(f"Column {i}: {X[:, i].dtype}")

        for col in X.columns:
            try:
                # Check if column can be converted to float
                X[col].astype(float)
                # If successful, convert the column
                X[col] = X[col].astype(float)
                # print(f"Column {col} converted to float")
            except ValueError:
                #do nothing
                pass
                # print(f"Column {col} contains non-float values")

        #################################################

        columns_df = []
        if isinstance(X, pd.DataFrame):
            columns_df = list(X.columns)
            if self.features_reduction is not None:
                columns_df = list(map(lambda x: f"Dimension_{x}", list(range(1, self.n_components + 1))))
            X = X.to_numpy()
        # Initialize missing values to their column means
        missing = ~np.isfinite(X)
        self.mu_ = np.nanmean(X, 0, keepdims=1)
        self.X_hat_ = np.where(missing, self.mu_, X)

        self.cls_ = None
        self.cls_features_reduction_ = None
        if self.features_reduction is not None or not np.any(missing):
            if self.features_reduction:
                assert self.features_reduction in ["umap", "pca"], "method need to be 'umap' or 'pca'"
                if self.features_reduction.lower() == "umap":
                    self.cls_features_reduction_ = umap.UMAP(n_components=self.n_components,
                                                             random_state=self.random_state,
                                                             n_jobs=self.n_jobs)
                elif self.features_reduction.lower() == "pca":
                    self.cls_features_reduction_ = PCA(n_components=self.n_components, random_state=self.random_state)

                self.cls_features_reduction_.fit(self.X_hat_)
                self.X_hat_ = self.cls_features_reduction_.transform(self.X_hat_)

            self.cls_ = KMeansConstrained(self.n_clusters,
                                          init="k-means++",
                                          size_min=self.size_min,
                                          random_state=self.random_state,
                                          n_jobs=self.n_jobs)

            self.labels_ = self.cls_.fit_predict(self.X_hat_)
            self.centroids_ = self.cls_.cluster_centers_

        else:
            prev_labels, labels = np.array([]), np.array([])
            prev_centroids, centroids = np.array([]), np.array([])
            for i in range(self.em_iter):
                if i > 0:
                    # initialize KMeans with the previous set of centroids. this is much
                    # faster and makes it easier to check convergence (since labels
                    # won't be permuted on every iteration), but might be more prone to
                    # getting stuck in local minima.
                    self.cls_ = KMeansConstrained(self.n_clusters,
                                                  init=prev_centroids,
                                                  size_min=self.size_min,
                                                  random_state=self.random_state,
                                                  n_jobs=self.n_jobs)
                else:
                    # do multiple random initializations in parallel
                    self.cls_ = KMeansConstrained(self.n_clusters,
                                                  init="k-means++",
                                                  size_min=self.size_min,
                                                  random_state=self.random_state,
                                                  n_jobs=self.n_jobs)

                # perform on the filled-in data
                self.labels_ = self.cls_.fit_predict(self.X_hat_)
                self.centroids_ = self.cls_.cluster_centers_

                # fill in the missing values based on their cluster centroids
                self.X_hat_[missing] = self.centroids_[self.labels_][missing]

                # when the labels have stopped changing then we have converged
                if i > 0 and np.all(self.labels_ == prev_labels):
                    break
                prev_labels = self.labels_
                prev_centroids = self.cls_.cluster_centers_

        self.inertia_ = self.cls_.inertia_
        self.cluster_centers_ = self.cls_.cluster_centers_

        if columns_df:
            self.X_hat_ = pd.DataFrame(self.X_hat_, columns=columns_df)

        return self

    def predict(self, X, size_min='init', size_max='init'):
        """
        Predict the closest cluster each sample in X belongs to given the provided constraints.
        The constraints can be temporally overridden when determining which cluster each datapoint is assigned to.

        Only computes the assignment step. It does not re-fit the cluster positions.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            New data to predict.

        size_min : int, optional, default: size_min provided with initialisation
            Constrain the label assignment so that each cluster has a minimum
            size of size_min. If None, no constrains will be applied.
            If 'init' the value provided during initialisation of the
            class will be used.

        size_max : int, optional, default: size_max provided with initialisation
            Constrain the label assignment so that each cluster has a maximum
            size of size_max. If None, no constrains will be applied.
            If 'init' the value provided during initialisation of the
            class will be used.

        Returns
        -------
        labels : array, shape [n_samples,]
            Index of the cluster each sample belongs to.
        """

        labels = self.cls_.predict(X, size_min=size_min, size_max=size_max)

        return labels

    def fit_predict(self, X, y=None):
        """Compute cluster centers and predict cluster index for each sample.

        Equivalent to calling fit(X) followed by predict(X) but also more efficient.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data to transform.

        Returns
        -------
        labels : array, shape [n_samples,]
            Index of the cluster each sample belongs to.
        """
        return self.fit(X).cls_.labels_


class KElbowVisualizer(ClusteringScoreVisualizer):
    """
    The K-Elbow Visualizer implements the "elbow" method of selecting the
    optimal number of clusters for K-means clustering. K-means is a simple
    unsupervised machine learning algorithm that groups data into a specified
    number (k) of clusters. Because the user must specify in advance what k to
    choose, the algorithm is somewhat naive -- it assigns all members to k
    clusters even if that is not the right k for the dataset.

    The elbow method runs k-means clustering on the dataset for a range of
    values for k (say from 1-10) and then for each value of k computes an
    average score for all clusters. By default, the ``distortion`` score is
    computed, the sum of square distances from each point to its assigned
    center. Other metrics can also be used such as the ``silhouette`` score,
    the mean silhouette coefficient for all samples or the
    ``calinski_harabasz`` score, which computes the ratio of dispersion between
    and within clusters.

    When these overall metrics for each model are plotted, it is possible to
    visually determine the best value for k. If the line chart looks like an
    arm, then the "elbow" (the point of inflection on the curve) is the best
    value of k. The "arm" can be either up or down, but if there is a strong
    inflection point, it is a good indication that the underlying model fits
    best at that point.

    .. note::
        | yellowbrick.cluster.elbow
        | Implements the elbow method for determining the optimal number of clusters.

        | Author:   Benjamin Bengfort
        | Created:  Thu Mar 23 22:36:31 2017 -0400

        | Copyright (C) 2016 The scikit-yb developers
        | For license information, see LICENSE.txt

        | ID: elbow.py [5a370c8] benjamin@bengfort.com

    Parameters
    ----------
    estimator : a scikit-learn clusterer
        Should be an instance of an unfitted clusterer, specifically ``KMeans`` or
        ``MiniBatchKMeans``. If it is not a clusterer, an exception is raised.

    ax : matplotlib Axes, default: None
        The axes to plot the figure on. If None is passed in the current axes
        will be used (or generated if required).

    k : integer, tuple, or iterable
        The k values to compute silhouette scores for. If a single integer
        is specified, then will compute the range (2,k). If a tuple of 2
        integers is specified, then k will be in np.arange(k[0], k[1]).
        Otherwise, specify an iterable of integers to use as values for k.

    metric : string, default: ``"distortion"``
        Select the scoring metric to evaluate the clusters. The default is the
        mean distortion, defined by the sum of squared distances between each
        observation and its closest centroid. Other metrics include:

        - **distortion**: mean sum of squared distances to centers
        - **silhouette**: mean ratio of intra-cluster and nearest-cluster distance
        - **calinski_harabasz**: ratio of within to between cluster dispersion

    timings : bool, default: True
        Display the fitting time per k to evaluate the amount of time required
        to train the clustering model.

    locate_elbow : bool, default: True
        Automatically find the "elbow" or "knee" which likely corresponds to the optimal
        value of k using the "knee point detection algorithm". The knee point detection
        algorithm finds the point of maximum curvature, which in a well-behaved
        clustering problem also represents the pivot of the elbow curve. The point is
        labeled with a dashed line and annotated with the score and k values.

    n_jobs : int, default: None
        Number of jobs to run in parallel. Training the estimator and computing
        the score are parallelized over the cross-validation splits.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.

    verbose : int, default: 0
        The verbosity level.

    pre_dispatch : int or str, default: '2*n_jobs'
        Controls the number of jobs that get dispatched during parallel
        execution. Reducing this number can be useful to avoid an
        explosion of memory consumption when more jobs get dispatched
        than CPUs can process. This parameter can be:
        - None, in which case all the jobs are immediately
        created and spawned. Use this for lightweight and
        fast-running jobs, to avoid delays due to on-demand
        spawning of the jobs
        - An int, giving the exact number of total jobs that are
        spawned
        - A str, giving an expression as a function of n_jobs,
        as in '2*n_jobs'

    kwargs : dict
        Keyword arguments that are passed to the base class and may influence
        the visualization as defined in other Visualizers.

    Attributes
    ----------
    k_scores\_ : array of shape (n,) where n is no. of k values
        The silhouette score corresponding to each k value.

    k_timers\_ : array of shape (n,) where n is no. of k values
        The time taken to fit n KMeans model corresponding to each k value.

    elbow_value\_ : integer
        The optimal value of k.

    elbow_score\_ : float
        The silhouette score corresponding to the optimal value of k.

    estimators\_ : ``BaseEstimator``
        a scikit-learn fitted estimator

    Examples
    --------
    >>> from yellowbrick.cluster import KElbowVisualizer
    >>> from sklearn.cluster import KMeans
    >>> model = KElbowVisualizer(KMeans(), k=10)
    >>> X = np.array([[1, 2], [1, 4], [1, 0],
    ...               [4, 2], [4, 4], [4, 0]])
    >>> model.fit(X)
    >>> model.show()

    Notes
    -----
    Modification from yellowbrick consist of get the best_estimator based
    on the finded elbow_value

    If you get a visualizer that doesn't have an elbow or inflection point,
    then this method may not be working. The elbow method does not work well
    if the data is not very clustered; in this case, you might see a smooth
    curve and the value of k is unclear. Other scoring methods, such as BIC or
    SSE, also can be used to explore if clustering is a correct choice.

    For a discussion on the Elbow method, read more at
    `Robert Gove's Block website <https://bl.ocks.org/rpgove/0060ff3b656618e9136b>`_.
    For more on the knee point detection algorithm see the paper `"Finding a "kneedle"
    in a Haystack" <https://raghavan.usc.edu//papers/kneedle-simplex11.pdf>`_.

    .. seealso::
        The scikit-learn documentation for the `silhouette_score
        <https://bit.ly/2LYWjYb>`_ and `calinski_harabasz_score
        <https://bit.ly/2ItAgts>`_. The default, ``distortion_score``, is
        implemented in ``yellowbrick.cluster.elbow``.
    """

    def __init__(
            self,
            estimator,
            ax=None,
            k=10,
            metric="distortion",
            timings=True,
            locate_elbow=True,
            n_jobs=None,
            verbose=0,
            pre_dispatch='2*n_jobs',
            **kwargs
    ):
        super(KElbowVisualizer, self).__init__(estimator, ax=ax, **kwargs)

        # Get the scoring method
        if metric not in KELBOW_SCOREMAP:
            raise YellowbrickValueError(
                "'{}' is not a defined metric "
                "use one of distortion, silhouette, or calinski_harabasz"
            )

        # Store the arguments
        self.scoring_metric = KELBOW_SCOREMAP[metric]
        self.metric = metric
        self.timings = timings
        self.locate_elbow = locate_elbow
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.pre_dispatch = pre_dispatch

        # Convert K into a tuple argument if an integer
        if isinstance(k, int):
            self.k_values_ = list(range(2, k + 1))
        elif (
                isinstance(k, tuple)
                and len(k) == 2
                and all(isinstance(x, (int, np.integer)) for x in k)
        ):
            self.k_values_ = list(range(*k))
        elif isinstance(k, Iterable) and all(
                isinstance(x, (int, np.integer)) for x in k
        ):
            self.k_values_ = list(k)
        else:
            raise YellowbrickValueError(
                (
                    "Specify an iterable of integers, a range, or maximal K value,"
                    " the value '{}' is not a valid argument for K.".format(k)
                )
            )

        # Holds the values of the silhoutte scores
        self.k_scores_ = None
        # Set Default Elbow Value
        self.elbow_value_ = None

    def fit(self, X, y=None, **kwargs):
        """
        Fits n KMeans models where n is the length of ``self.k_values_``,
        storing the silhouette scores in the ``self.k_scores_`` attribute.
        The "elbow" and silhouette score corresponding to it are stored in
        ``self.elbow_value`` and ``self.elbow_score`` respectively.
        This method finishes up by calling draw to create the plot.
        """

        self.k_scores_ = []
        self.k_timers_ = []
        self.estimators_ = []

        # We clone the estimator to make sure that all the folds are
        # independent, and that it is pickle-able.
        parallel = Parallel(n_jobs=self.n_jobs, verbose=self.verbose,
                            pre_dispatch=self.pre_dispatch)
        results = parallel(
            delayed(_fit_and_score)(
                clone(self.estimator), X, k, self.scoring_metric, **kwargs)
            for k in self.k_values_)

        self.k_scores_, self.k_timers_, self.estimators_ = zip(*results)

        if self.locate_elbow:
            locator_kwargs = {
                "distortion": {
                    "curve": "convex",
                    "direction": "decreasing",
                },
                "silhouette": {
                    "curve": "concave",
                    "direction": "increasing",
                },
                "calinski_harabasz": {
                    "curve": "concave",
                    "direction": "increasing",
                },
            }.get(self.metric, {})
            elbow_locator = KneeLocator(
                self.k_values_, self.k_scores_, **locator_kwargs
            )
            if elbow_locator.knee is None:
                self.elbow_value_ = None
                self.elbow_score_ = 0
                warning_message = (
                    "No 'knee' or 'elbow' point detected, "
                    "pass `locate_elbow=False` to remove the warning"
                )
                warnings.warn(warning_message, YellowbrickWarning)
            else:
                self.elbow_value_ = elbow_locator.knee
                self.elbow_score_ = self.k_scores_[self.k_values_.index(self.elbow_value_)]
                self.best_estimator_ = self.estimators_[self.k_values_.index(self.elbow_value_)]

        self.draw()

        return self

    def draw(self):
        """
        Draw the elbow curve for the specified scores and values of K.
        """
        # Plot the silhouette score against k
        self.ax.plot(self.k_values_, self.k_scores_, marker="D")
        if self.locate_elbow is True and self.elbow_value_ is not None:
            elbow_label = "elbow at $k={}$, $score={:0.3f}$".format(
                self.elbow_value_, self.elbow_score_
            )
            self.ax.axvline(
                self.elbow_value_, c=LINE_COLOR, linestyle="--", label=elbow_label
            )

        # If we're going to plot the timings, create a twinx axis
        if self.timings:
            self.axes = [self.ax, self.ax.twinx()]
            self.axes[1].plot(
                self.k_values_,
                self.k_timers_,
                label="fit time",
                c="g",
                marker="o",
                linestyle="--",
                alpha=0.75,
            )

        return self.ax

    def finalize(self):
        """
        Prepare the figure for rendering by setting the title as well as the
        X and Y axis labels and adding the legend.

        """
        # Get the metric name
        metric = self.scoring_metric.__name__.replace("_", " ").title()

        # Set the title
        self.set_title("{} Elbow for {} Clustering".format(metric, self.name))

        # Set the x and y labels
        self.ax.set_xlabel("k")
        self.ax.set_ylabel(metric.lower())

        # set the legend if locate_elbow=True
        if self.locate_elbow is True and self.elbow_value_ is not None:
            self.ax.legend(loc="best", fontsize="medium", frameon=True)

        # Set the second y axis labels
        if self.timings:
            self.axes[1].grid(False)
            self.axes[1].set_ylabel("fit time (seconds)", color="g")
            self.axes[1].tick_params("y", colors="g")


def _fit_and_score(estimator, X, n_clusters, scoring_metric, **kwargs):
    """
    Fit estimator and compute scores for a given dataset split.

    Parameters
    ----------
    estimator : estimator object implementing 'fit'
        The object to use to fit the data.
    X : array-like of shape (n_samples, n_features)
        The data to fit.
    n_clusters : int,
        The number of clusters (k) to form as well as the number of centroids to generate.
    scoring_metric : scoring_metric of KElbow

    Returns
    -------
    score : The silhouette score corresponding to the n_clusters.

    timers : The time taken to fit n KMeans model.

    estimator : a scikit-learn fitted clusterer
    """
    start = time.time()
    # Set the k value and fit the model
    estimator.set_params(n_clusters=n_clusters)
    estimator.fit(X, **kwargs)
    # Append the time and score to our plottable metrics
    timers = time.time() - start
    # get X_hat_ return from fit if exist
    # avoid: ValueError: Input contains NaN, infinity or a value too large for dtype('float64').
    score = scoring_metric(X if not hasattr(estimator, "X_hat_") else estimator.X_hat_, estimator.labels_)
    return score, timers, estimator,


def optimal_clustering(X: Union[pd.DataFrame, np.ndarray],
                       size_min: int = 10,
                       metric: str = "distortion",
                       features_reduction: Optional[str] = None,
                       n_components: int = 2,
                       n_jobs: int = 1,
                       random_state: Optional[int] = None,
                       visualize: bool = False) \
        -> Tuple[KMeansConstrained, Union[umap.UMAP, PCA], int, np.ndarray, int, Sequence[
            float], float, np.ndarray, np.ndarray, np.ndarray]:
    """
    Function to find the optimal clustering using a constrained k means. Two method are available to find the optimal
    number of cluster ``silhouette`` or ``elbow``.

    :param X: array-like or DataFrame of floats, shape (n_samples, n_features)
        The observations to cluster.
    :param size_min: Constrain the label assignment so that each cluster has a minimum size of size_min.
        If None, no constrains will be applied. default: None
    :param metric: Select the scoring metric to evaluate the clusters.
        The default is the mean distortion, defined by the sum of squared distances between each observation
        and its closest centroid. Other metrics include:
        - distortion: mean sum of squared distances to centers
        - silhouette: mean ratio of intra-cluster and nearest-cluster distance
        - calinski_harabasz: ratio of within to between cluster dispersion
    :param features_reduction: Method for reduction of the embedded space with n_components. Can be pca or umap.
        Default: None
    :param n_components: Dimension of the embedded space for features reduction. Default 2.
    :param n_jobs: int
        The number of jobs to use for the computation. This works by computing
        each of the n_init runs in parallel.
        If -1 all CPUs are used. If 1 is given, no parallel computing code is
        used at all, which is useful for debugging. For n_jobs below -1,
        (n_cpus + 1 + n_jobs) are used. Thus for n_jobs = -2, all CPUs but one
        are used.
    :param random_state: int, RandomState instance or None, optional, default: None
        If int, random_state is the seed used by the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.
    :param visualize: bool, default: False
        If True, calls ``show()``
    :return:
        - cls:
          KMeansConstrained classifier object
        - cls_features_reduction:
          PCA or UMAP reduction object
        - cluster_nb:
          optimal number of cluster
        - labels:
          label[i] is the code or index of the centroid the i'th observation is closest to.
        - ref_label:
          cluster label with the minimal within-cluster sum-of-squares.
        - wicss_clusters:
          within-cluster sum-of-squares for each cluster
        - best_wicss_cluster:
          minimal wicss.
        - centroid:
          Centroids found at the last iteration of k-means.
        - X_hat:
          Copy of X with the missing values filled in.
    """
    assert metric in ["distortion", "silhouette", "calinski_harabasz"], "method need to be " \
                                                                        "'distortion' or " \
                                                                        "'silhouette' or " \
                                                                        "'calinski_harabasz'"

    n_samples = X.shape[0]
    min_cluster = 2 if metric in ["silhouette", "calinski_harabasz"] else 1
    max_cluster = n_samples // size_min + 1

    K = range(min_cluster, max_cluster)

    assert K, ValueError(f"Number of cluster is incompatible with metric {metric}, min cluster is {min_cluster} and "
                         f"max cluster is {max_cluster - 1} with a size_min of {size_min}")

    if max_cluster - 1 == 1:
        warnings.warn("Only one cluster is possible", RuntimeWarning, stacklevel=2)

    elbow_visualizer = KElbowVisualizer(estimator=KMeansConstrainedMissing(size_min=size_min,
                                                                           em_iter=10,
                                                                           features_reduction=features_reduction,
                                                                           n_components=n_components,
                                                                           n_jobs=1,
                                                                           random_state=random_state),
                                        k=K,
                                        metric=metric,
                                        locate_elbow=True,
                                        n_jobs=n_jobs)

    # fit
    elbow_visualizer.fit(X)

    if visualize:
        elbow_visualizer.show()

    cluster_nb = elbow_visualizer.elbow_value_

    # get cls, labels, centroids, inertia, Xhat corresponding to cluster_nb
    if cluster_nb == 1:
        warnings.warn("Only one cluster")
    if not cluster_nb:
        raise ValueError("Unable to determine a number of cluster")

    cls = elbow_visualizer.best_estimator_
    cls_features_reduction = elbow_visualizer.best_estimator_.cls_features_reduction_
    labels = elbow_visualizer.best_estimator_.labels_
    centroids = elbow_visualizer.best_estimator_.centroids_
    X_hat = elbow_visualizer.best_estimator_.X_hat_
    mu = elbow_visualizer.best_estimator_.mu_

    # cluster choose for reference is the one which minimizing a criterion known as the inertia
    # or within-cluster sum-of-squares (WCSS)
    # WCSS is defined as the sum of the squared distance between each member of the cluster and its centroid.
    # WSS=\sum ^{M}_{i=1}\left( x_{i}-c_{i}\right) ^{2}

    # Calculate the distances matrix between all data points and the final centroids.
    D = cdist(X_hat, centroids, 'euclidean')
    wicss_clusters = []
    for label in range(0, cluster_nb):
        labels_bool = labels == label
        # distance to the closest centroid
        dist = D[labels_bool, label]
        # Total with-in sum of square
        wicss_cluster = sum(dist ** 2)
        wicss_clusters.append(wicss_cluster)
        # best is minimal wicss

    best_wicss_cluster = np.min(wicss_clusters)
    ref_label = np.argmin(wicss_clusters)

    return cls, cls_features_reduction, cluster_nb, labels, ref_label, wicss_clusters, best_wicss_cluster, centroids, X_hat, mu
